@@ -15,8 +15,9 @@
  * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
+#include <weight.h>
 #include <lrx_compiler.h>
-
+#include <cstdint>
 
 using namespace std;
 
@@ -33,6 +34,9 @@ wstring const LRXCompiler::LRX_COMPILER_REPEAT_ELEM     = L"repeat";
 wstring const LRXCompiler::LRX_COMPILER_SEQ_ELEM        = L"seq";
 
 wstring const LRXCompiler::LRX_COMPILER_LEMMA_ATTR      = L"lemma";
+wstring const LRXCompiler::LRX_COMPILER_SUFFIX_ATTR     = L"suffix";
+wstring const LRXCompiler::LRX_COMPILER_CONTAINS_ATTR   = L"contains";
+wstring const LRXCompiler::LRX_COMPILER_CASE_ATTR       = L"case";
 wstring const LRXCompiler::LRX_COMPILER_SURFACE_ATTR    = L"surface";
 wstring const LRXCompiler::LRX_COMPILER_TAGS_ATTR       = L"tags";
 wstring const LRXCompiler::LRX_COMPILER_WEIGHT_ATTR     = L"weight";
@@ -101,6 +105,8 @@ LRXCompiler::LRXCompiler()
 
   alphabet.includeSymbol(L"<ANY_TAG>");
   alphabet.includeSymbol(L"<ANY_CHAR>");
+  alphabet.includeSymbol(L"<ANY_UPPER>");
+  alphabet.includeSymbol(L"<ANY_LOWER>");
   alphabet.includeSymbol(L"<$>");
 
 }
@@ -148,6 +154,27 @@ LRXCompiler::attrib(wstring const &name)
   return XMLParseUtil::attrib(reader, name);
 }
 
+wstring
+LRXCompiler::attrib(wstring const &name, const wstring fallback)
+{
+  string mystr = "";
+  for (int i = 0, limit = name.size(); i != limit; i++) {
+    mystr += static_cast<char>(name[i]);
+  }
+
+  xmlChar *attrname = xmlCharStrdup(mystr.c_str());
+  xmlChar *myattr = xmlTextReaderGetAttribute(reader, attrname);
+  wstring result = XMLParseUtil::towstring(myattr);
+  xmlFree(myattr);
+  xmlFree(attrname);
+  if(myattr == NULL) {
+    return fallback;
+  }
+  else {
+    return result;
+  }
+}
+
 bool
 LRXCompiler::allBlanks()
 {
@@ -262,8 +289,8 @@ LRXCompiler::procRule()
     int ret = xmlTextReaderRead(reader);
     if(ret != 1)
     {
-      cerr << L"Error (" << xmlTextReaderGetParserLineNumber(reader);
-      cerr << L"): Parse error." << endl;
+      wcerr << L"Error (" << xmlTextReaderGetParserLineNumber(reader);
+      wcerr << L"): Parse error." << endl;
       exit(EXIT_FAILURE);
     }
 
@@ -321,14 +348,15 @@ LRXCompiler::procOr()
     fwprintf(stderr, L"    or: \n");
   }
 
+  int or_initial_state = currentState;
   vector<int> reachedStates;
   while(true)
   {
     int ret = xmlTextReaderRead(reader);
     if(ret != 1)
     {
-      cerr << L"Error (" << xmlTextReaderGetParserLineNumber(reader);
-      cerr << L"): Parse error." << endl;
+      wcerr << L"Error (" << xmlTextReaderGetParserLineNumber(reader);
+      wcerr << L"): Parse error." << endl;
       exit(EXIT_FAILURE);
     }
 
@@ -337,13 +365,13 @@ LRXCompiler::procOr()
 
     if(name == LRX_COMPILER_MATCH_ELEM)
     {
-      currentState = lastState;
+      currentState = transducer.insertNewSingleTransduction(0, or_initial_state);
       procMatch();
       reachedStates.push_back(currentState);
     }
     else if(name == LRX_COMPILER_SEQ_ELEM)
     {
-      currentState = lastState;
+      currentState = transducer.insertNewSingleTransduction(0, or_initial_state);
       procSeq();
       reachedStates.push_back(currentState);
     }
@@ -351,13 +379,13 @@ LRXCompiler::procOr()
     {
       if(reachedStates.size() > 1)
       {
-        for(vector<int>::iterator it = reachedStates.begin(); it != reachedStates.end(); it++)
+        for(auto& it : reachedStates)
         {
-          if(*it == currentState)
+          if(it == currentState)
           {
             continue;
           }
-          transducer.linkStates(*it, currentState, 0);
+          transducer.linkStates(it, currentState, 0);
         }
       }
       break;
@@ -390,8 +418,8 @@ LRXCompiler::procDefSeq()
     int ret = xmlTextReaderRead(reader);
     if(ret != 1)
     {
-      cerr << L"Error (" << xmlTextReaderGetParserLineNumber(reader);
-      cerr << L"): Parse error." << endl;
+      wcerr << L"Error (" << xmlTextReaderGetParserLineNumber(reader);
+      wcerr << L"): Parse error." << endl;
       exit(EXIT_FAILURE);
     }
 
@@ -411,6 +439,10 @@ LRXCompiler::procDefSeq()
     {
       procRepeat();
     }
+    else if(name == LRX_COMPILER_SEQ_ELEM)
+    {
+      procSeq();
+    }
     else if(name == LRX_COMPILER_DEFSEQ_ELEM)
     {
       transducer.setFinal(currentState);
@@ -435,10 +467,17 @@ LRXCompiler::procDefSeq()
 void
 LRXCompiler::procMatch()
 {
-  wstring lemma = this->attrib(LRX_COMPILER_LEMMA_ATTR);
-  wstring tags = this->attrib(LRX_COMPILER_TAGS_ATTR);
+  // These are mutually exclusive
+  wstring lemma = this->attrib(LRX_COMPILER_LEMMA_ATTR, L"*");
+  wstring contains = this->attrib(LRX_COMPILER_SUFFIX_ATTR);
+  wstring suffix = this->attrib(LRX_COMPILER_CONTAINS_ATTR);
+  wstring _case = this->attrib(LRX_COMPILER_CASE_ATTR); // This could potentially be non-exclusive
+
+  // This is currently disabled: Future use
   wstring surface = this->attrib(LRX_COMPILER_SURFACE_ATTR);
 
+  wstring tags = this->attrib(LRX_COMPILER_TAGS_ATTR, L"*");
+
   if(surface != L"")
   {
     if(debugMode)
@@ -446,25 +485,43 @@ LRXCompiler::procMatch()
       fwprintf(stderr, L"      match: %S\n", surface.c_str());
     }
 
-    for(wstring::iterator it = surface.begin(); it != surface.end(); it++)
+    for(auto& it : surface)
     {
-      currentState = transducer.insertSingleTransduction(alphabet(*it, 0), currentState);
+      currentState = transducer.insertSingleTransduction(alphabet(it, 0), currentState);
     }
   }
   else
   {
     if(debugMode)
     {
-      fwprintf(stderr, L"      match: %S, %S\n", lemma.c_str(), tags.c_str());
+      fwprintf(stderr, L"      match: [%S, %S, %S, %S] %S\n", lemma.c_str(), suffix.c_str(), contains.c_str(), _case.c_str(), tags.c_str());
     }
 
-
-    if(lemma == L"*") {
-      lemma = L"";
-    }
-
-    if(lemma == L"")
+    if(_case != L"")
     {
+      if(_case == L"AA") // <ANY_UPPER>+
+      {
+        int localLast = currentState;
+        currentState = transducer.insertSingleTransduction(alphabet(alphabet(L"<ANY_UPPER>"), 0), currentState);
+        transducer.linkStates(currentState, localLast, 0);
+      }
+      else if(_case == L"aa")  // <ANY_LOWER>+
+      {
+        int localLast = currentState;
+        currentState = transducer.insertSingleTransduction(alphabet(alphabet(L"<ANY_LOWER>"), 0), currentState);
+        transducer.linkStates(currentState, localLast, 0);
+      }
+      else if(_case == L"Aa") // <ANY_UPPER>+ <ANY_LOWER>+
+      {
+        currentState = transducer.insertSingleTransduction(alphabet(alphabet(L"<ANY_UPPER>"), 0), currentState);
+        int localLast = currentState;
+        currentState = transducer.insertSingleTransduction(alphabet(alphabet(L"<ANY_LOWER>"), 0), currentState);
+        transducer.linkStates(currentState, localLast, 0);
+      }
+    }
+    if(lemma == L"*" && suffix == L"" && contains == L"" && _case == L"")
+    {
+      // This is only if there is no suffix or case or contains
       if(debugMode)
       {
         fwprintf(stderr, L"        char: -\n");
@@ -473,57 +530,52 @@ LRXCompiler::procMatch()
       currentState = transducer.insertSingleTransduction(alphabet(alphabet(L"<ANY_CHAR>"), 0), currentState);
       transducer.linkStates(currentState, localLast, 0);
     }
-    else
+    else if(suffix != L"")
     {
-      for(wstring::iterator it = lemma.begin(); it != lemma.end(); it++)
+      // A suffix is <ANY_CHAR> any amount of times followed by whatever is in the suffix
+      int localLast = currentState;
+      currentState = transducer.insertSingleTransduction(alphabet(alphabet(L"<ANY_CHAR>"), 0), currentState);
+      transducer.linkStates(currentState, localLast, 0);
+      for(auto& it : suffix)
       {
-        currentState = transducer.insertSingleTransduction(alphabet(*it, 0), currentState);
+        currentState = transducer.insertSingleTransduction(alphabet(it, 0), currentState);
       }
     }
-
-    if(tags != L"")
+    else if(contains != L"")
     {
-      wstring tag = L"";
-      for(wstring::iterator it = tags.begin(); it != tags.end(); it++)
+      // A contains is <ANY_CHAR> any amount of times followed by whatever is in the attribute
+      // followed by <ANY_CHAR> any amount of times
+      int localLast = currentState;
+      currentState = transducer.insertSingleTransduction(alphabet(alphabet(L"<ANY_CHAR>"), 0), currentState);
+      transducer.linkStates(currentState, localLast, 0);
+      for(auto& it : suffix)
       {
-        if(*it == L'.')
+        currentState = transducer.insertSingleTransduction(alphabet(it, 0), currentState);
+      }
+      currentState = transducer.insertSingleTransduction(alphabet(alphabet(L"<ANY_CHAR>"), 0), currentState);
+      transducer.linkStates(currentState, localLast, 0);
+    }
+    else if(lemma != L"*")
+    {
+      for(auto& it : lemma)
+      {
+        currentState = transducer.insertSingleTransduction(alphabet(it, 0), currentState);
+      }
+    }
+    else
+    {
+      fwprintf(stderr, L"Something surprising happened in <match> compilation\n");
+    }
+
+    wstring tag = L"";
+    for(auto& it : tags)
+    {
+      if(it == L'.')
+      {
+        if(tag == L"")
         {
-          tag = L"<" + tag + L">";
-          if(!alphabet.isSymbolDefined(tag.c_str()))
-          {
-            alphabet.includeSymbol(tag.c_str());
-          }
-          if(debugMode)
-          {
-            fwprintf(stderr, L"        tag: %S\n", tag.c_str());
-          }
-          if(tag == L"<*>")
-          {
-	    int localLast = currentState;
-            currentState = transducer.insertSingleTransduction(alphabet(alphabet(L"<ANY_TAG>"), 0), currentState);
-	    transducer.linkStates(currentState, localLast, 0);
-          }
-          else
-          {
-            currentState = transducer.insertSingleTransduction(alphabet(alphabet(tag.c_str()), 0), currentState);
-          }
-          tag = L"";
           continue;
         }
-        tag = tag + *it;
-      }
-      if(tag == L"*")
-      {
-        if(debugMode)
-        {
-          fwprintf(stderr, L"        tag: %S\n", tag.c_str());
-        }
-        int localLast = currentState;
-        currentState = transducer.insertSingleTransduction(alphabet(alphabet(L"<ANY_TAG>"), 0), currentState);
-        transducer.linkStates(currentState, localLast, 0);
-      }
-      else
-      {
         tag = L"<" + tag + L">";
         if(!alphabet.isSymbolDefined(tag.c_str()))
         {
@@ -533,19 +585,47 @@ LRXCompiler::procMatch()
         {
           fwprintf(stderr, L"        tag: %S\n", tag.c_str());
         }
-        currentState = transducer.insertSingleTransduction(alphabet(alphabet(tag.c_str()), 0), currentState);
+        if(tag == L"<*>")
+        {
+          int localLast = currentState;
+          currentState = transducer.insertSingleTransduction(alphabet(alphabet(L"<ANY_TAG>"), 0), currentState);
+          transducer.linkStates(currentState, localLast, 0);
+        }
+        else
+        {
+          currentState = transducer.insertSingleTransduction(alphabet(alphabet(tag.c_str()), 0), currentState);
+        }
+        tag = L"";
+        continue;
       }
+      tag = tag + it;
     }
-    else
+    if(tag == L"*")
     {
       if(debugMode)
       {
-        fwprintf(stderr, L"        tag: -\n");
+        fwprintf(stderr, L"        tag: %S\n", tag.c_str());
       }
       int localLast = currentState;
       currentState = transducer.insertSingleTransduction(alphabet(alphabet(L"<ANY_TAG>"), 0), currentState);
       transducer.linkStates(currentState, localLast, 0);
     }
+    else if(tag == L"")
+    {
+    }
+    else
+    {
+      tag = L"<" + tag + L">";
+      if(!alphabet.isSymbolDefined(tag.c_str()))
+      {
+        alphabet.includeSymbol(tag.c_str());
+      }
+      if(debugMode)
+      {
+        fwprintf(stderr, L"        tag: %S\n", tag.c_str());
+      }
+      currentState = transducer.insertSingleTransduction(alphabet(alphabet(tag.c_str()), 0), currentState);
+    }
   }
 
   if(xmlTextReaderIsEmptyElement(reader))
@@ -562,8 +642,8 @@ LRXCompiler::procMatch()
     int ret = xmlTextReaderRead(reader);
     if(ret != 1)
     {
-      cerr << L"Error (" << xmlTextReaderGetParserLineNumber(reader);
-      cerr << L"): Parse error." << endl;
+      wcerr << L"Error (" << xmlTextReaderGetParserLineNumber(reader);
+      wcerr << L"): Parse error." << endl;
       exit(EXIT_FAILURE);
     }
 
@@ -574,8 +654,8 @@ LRXCompiler::procMatch()
     {
       if(!canSelect)
       {
-        cerr << L"Error (" << xmlTextReaderGetParserLineNumber(reader);
-        cerr << L"): <select> is not permitted inside <repeat>." << endl;
+        wcerr << L"Error (" << xmlTextReaderGetParserLineNumber(reader);
+        wcerr << L"): <select> is not permitted inside <repeat>." << endl;
         exit(EXIT_FAILURE);
       }
       procSelect();
@@ -584,8 +664,8 @@ LRXCompiler::procMatch()
     {
       if(!canSelect)
       {
-        cerr << L"Error (" << xmlTextReaderGetParserLineNumber(reader);
-        cerr << L"): <remove> is not permitted inside <repeat>." << endl;
+        wcerr << L"Error (" << xmlTextReaderGetParserLineNumber(reader);
+        wcerr << L"): <remove> is not permitted inside <repeat>." << endl;
         exit(EXIT_FAILURE);
       }
       procRemove();
@@ -611,14 +691,9 @@ void
 LRXCompiler::procSelect()
 {
 
-  wstring lemma =this->attrib(LRX_COMPILER_LEMMA_ATTR);
+  wstring lemma =this->attrib(LRX_COMPILER_LEMMA_ATTR, L"*");
   wstring tags =this->attrib(LRX_COMPILER_TAGS_ATTR);
 
-  if(lemma == L"*")
-  {
-    lemma = L"";
-  }
-
   wstring key = L"<" + LRX_COMPILER_TYPE_SELECT + L">" + lemma;
 
   Transducer recogniser;
@@ -633,7 +708,7 @@ LRXCompiler::procSelect()
   currentState = transducer.insertSingleTransduction(alphabet(0, alphabet(L"<" + LRX_COMPILER_TYPE_SELECT + L">")), currentState);
 
 
-  if(lemma == L"")
+  if(lemma == L"*")
   {
     currentState = transducer.insertSingleTransduction(alphabet(0, alphabet(L"<ANY_CHAR>")), currentState);
     int localLast = localCurrentState;
@@ -642,18 +717,18 @@ LRXCompiler::procSelect()
     key = key + L"<ANY_CHAR>";
   }
 
-  for(wstring::iterator it = lemma.begin(); it != lemma.end(); it++)
+  for(auto& it : lemma)
   {
-    currentState = transducer.insertSingleTransduction(alphabet(0, *it), currentState);
-    localCurrentState = recogniser.insertSingleTransduction(alphabet(*it, 0), localCurrentState);
+    currentState = transducer.insertSingleTransduction(alphabet(0, it), currentState);
+    localCurrentState = recogniser.insertSingleTransduction(alphabet(it, 0), localCurrentState);
   }
 
   if(tags != L"")
   {
     wstring tag = L"";
-    for(wstring::iterator it = tags.begin(); it != tags.end(); it++)
+    for(auto& it : tags)
     {
-      if(*it == L'.')
+      if(it == L'.')
       {
         tag = L"<" + tag + L">";
         if(!alphabet.isSymbolDefined(tag.c_str()))
@@ -681,7 +756,7 @@ LRXCompiler::procSelect()
         tag = L"";
         continue;
       }
-      tag = tag + *it;
+      tag = tag + it;
     }
     if(tag == L"*")
     {
@@ -741,16 +816,15 @@ void
 LRXCompiler::procRemove()
 {
 
-  wstring lemma =this->attrib(LRX_COMPILER_LEMMA_ATTR);
+  wstring lemma =this->attrib(LRX_COMPILER_LEMMA_ATTR, L"*");
   wstring tags =this->attrib(LRX_COMPILER_TAGS_ATTR);
 
-  if(lemma == L"*")
+  wstring key = L"<" + LRX_COMPILER_TYPE_REMOVE + L">";
+  if(lemma != L"*")
   {
-    lemma = L"";
+    key += lemma;
   }
 
-  wstring key = L"<" + LRX_COMPILER_TYPE_REMOVE + L">" + lemma;
-
   Transducer recogniser;
   int localCurrentState = recogniser.getInitial();
 
@@ -762,7 +836,7 @@ LRXCompiler::procRemove()
   currentState = transducer.insertSingleTransduction(alphabet(alphabet(L"<$>"), alphabet(L"<$>")), currentState);
   currentState = transducer.insertSingleTransduction(alphabet(0, alphabet(L"<" + LRX_COMPILER_TYPE_REMOVE + L">")), currentState);
 
-  if(lemma == L"")
+  if(lemma == L"*")
   {
     currentState = transducer.insertSingleTransduction(alphabet(0, alphabet(L"<ANY_CHAR>")), currentState);
     int localLast = localCurrentState;
@@ -770,19 +844,21 @@ LRXCompiler::procRemove()
     recogniser.linkStates(localCurrentState, localLast, 0);
     key = key + L"<ANY_CHAR>";
   }
-
-  for(wstring::iterator it = lemma.begin(); it != lemma.end(); it++)
+  else
   {
-    currentState = transducer.insertSingleTransduction(alphabet(0, *it), currentState);
-    localCurrentState = recogniser.insertSingleTransduction(alphabet(*it, 0), localCurrentState);
+    for(auto& it : lemma)
+    {
+      currentState = transducer.insertSingleTransduction(alphabet(0, it), currentState);
+      localCurrentState = recogniser.insertSingleTransduction(alphabet(it, 0), localCurrentState);
+    }
   }
 
   if(tags != L"")
   {
     wstring tag = L"";
-    for(wstring::iterator it = tags.begin(); it != tags.end(); it++)
+    for(auto& it : tags)
     {
-      if(*it == L'.')
+      if(it == L'.')
       {
         tag = L"<" + tag + L">";
         if(!alphabet.isSymbolDefined(tag.c_str()))
@@ -810,7 +886,7 @@ LRXCompiler::procRemove()
         tag = L"";
         continue;
       }
-      tag = tag + *it;
+      tag = tag + it;
     }
     if(tag == L"*")
     {
@@ -877,14 +953,14 @@ LRXCompiler::procRepeat()
   int upto = stoi(xupto);
   if(from < 0 || upto < 0)
   {
-    cerr << L"Error (" << xmlTextReaderGetParserLineNumber(reader);
-    cerr << L"): Number of repetitions cannot be negative." << endl;
+    wcerr << L"Error (" << xmlTextReaderGetParserLineNumber(reader);
+    wcerr << L"): Number of repetitions cannot be negative." << endl;
     exit(EXIT_FAILURE);
   }
   else if(from > upto)
   {
-    cerr << L"Error (" << xmlTextReaderGetParserLineNumber(reader);
-    cerr << L"): Lower bound on number of repetitions cannot be larger than upper bound." << endl;
+    wcerr << L"Error (" << xmlTextReaderGetParserLineNumber(reader);
+    wcerr << L"): Lower bound on number of repetitions cannot be larger than upper bound." << endl;
     exit(EXIT_FAILURE);
   }
   int count = upto - from;
@@ -898,8 +974,8 @@ LRXCompiler::procRepeat()
     int ret = xmlTextReaderRead(reader);
     if(ret != 1)
     {
-      cerr << L"Error (" << xmlTextReaderGetParserLineNumber(reader);
-      cerr << L"): Parse error." << endl;
+      wcerr << L"Error (" << xmlTextReaderGetParserLineNumber(reader);
+      wcerr << L"): Parse error." << endl;
       exit(EXIT_FAILURE);
     }
 
@@ -968,36 +1044,32 @@ LRXCompiler::write(FILE *fst)
   alphabet.write(fst);
 
   Compression::multibyte_write(recognisers.size(), fst);
-  for(map<wstring, Transducer>::iterator it = recognisers.begin(); it != recognisers.end(); it++)
+  for(auto& it : recognisers)
   {
-    Compression::wstring_write(it->first, fst);
+    Compression::wstring_write(it.first, fst);
     if(debugMode)
     {
-      fwprintf(stderr, L"+ %d => %S\n", it->second.size(), it->first.c_str());
-      it->second.show(alphabet, stderr);
+      fwprintf(stderr, L"+ %d => %S\n", it.second.size(), it.first.c_str());
+      it.second.show(alphabet, stderr, 0, false);
     }
-    it->second.write(fst);
+    it.second.write(fst);
   }
 
   Compression::wstring_write(L"main", fst);
   if(outputGraph)
   {
-    transducer.show(alphabet, stderr);
+    transducer.show(alphabet, stderr, 0, false);
   }
   transducer.write(fst);
 
-  struct weight {
-        int id;
-        double pisu;
-  };
-
-  for(map<int, double>::iterator it = weights.begin(); it != weights.end(); it++)
+  for(auto& it : weights)
   {
     if(debugMode)
     {
-      fwprintf(stderr, L"%.4f %d\n", it->second, it->first);
+      fwprintf(stderr, L"%.4f %d\n", it.second, it.first);
     }
-    weight record = {it->first, it->second};
+    weight record{it.first, "", it.second};
+    weight_to_le(record);
     fwrite((void *)&record, 1, sizeof(weight), fst);
   }
 
