@@ -15,8 +15,9 @@
  * along with this program; if not, see <http://www.gnu.org/licenses/>.
  */
 
+#include <weight.h>
 #include <lrx_processor.h>
-
+#include <cstdint>
 using namespace std;
 
 wstring const LRXProcessor::LRX_PROCESSOR_TAG_SELECT     = L"<select>";
@@ -89,6 +90,10 @@ LRXProcessor::load(FILE *in)
       len2--;
     }
     recognisers[name].read(in, alphabet);
+    if(debugMode)
+    {
+      fwprintf(stderr, L"Recogniser: %S, [finals: %d]\n", name.c_str(), recognisers[name].getFinals().size());
+    }
     len--;
   }
 
@@ -109,16 +114,10 @@ LRXProcessor::load(FILE *in)
   transducer.read(in, alphabet);
 
   // Now read in weights
-  struct weight {
-        int id;
-        double pisu;
-  };
-
-  while(!feof(in))
+  weight record;
+  while(fread(&record, sizeof(weight), 1, in))
   {
-    weight record;
-    fread(&record, sizeof(weight), 1, in);
-
+    weight_from_le(record);
     wstring sid = L"<" + itow(record.id) + L">";
     weights[sid] = record.pisu;
 
@@ -126,7 +125,6 @@ LRXProcessor::load(FILE *in)
     {
       //fwprintf(stderr, L"%S %d weight(%.4f)\n", sid.c_str(), record.id, record.pisu);
     }
-
   }
 
   return;
@@ -172,14 +170,9 @@ LRXProcessor::readFullBlock(FILE *input, wchar_t const delim1, wchar_t const del
 bool
 LRXProcessor::recognisePattern(const wstring lu, const wstring op)
 {
-  if(debugMode)
-  {
-    fwprintf(stderr, L"================================================\n");
-  }
-
   if(recognisers.count(op) < 1)
   {
-    fwprintf(stderr, L"WARNING: Recogniser size 0 for key %S, skipping...\n", op.c_str());
+    fwprintf(stderr, L"WARNING: Recogniser not found for key %S, skipping... [LU: %S]\n", op.c_str(), lu.c_str());
     return false;
   }
 
@@ -193,7 +186,7 @@ LRXProcessor::recognisePattern(const wstring lu, const wstring op)
   bool readingTag = false;
   wstring tag = L"";
   int val = 0;
-  for(wstring::const_iterator it = lu.begin(); it != lu.end(); it++)
+  for(auto& it : lu)
   {
 /*
     if(debugMode)
@@ -206,16 +199,16 @@ LRXProcessor::recognisePattern(const wstring lu, const wstring op)
     {
       return false;
     }
-    if(*it == L'<')
+    if(it == L'<')
     {
       tag = L"";
       readingTag = true;
-      tag = tag + *it;
+      tag = tag + it;
       continue;
     }
-    if(*it == L'>')
+    if(it == L'>')
     {
-      tag = tag + *it;
+      tag = tag + it;
       val = static_cast<int>(alphabet(tag));
       if(val == 0)
       {
@@ -234,11 +227,12 @@ LRXProcessor::recognisePattern(const wstring lu, const wstring op)
     }
     if(readingTag)
     {
-      tag = tag + *it;
+      tag = tag + it;
     }
     else
     {
-      int val = static_cast<int>(*it);
+      // We're not in a tag, so were just reading characters
+      int val = static_cast<int>(it);
 /*
       if(debugMode)
       {
@@ -247,14 +241,19 @@ LRXProcessor::recognisePattern(const wstring lu, const wstring op)
 */
       //cur.step(val, a(L"<ANY_CHAR>"));
       //cur.step(val);
+      set<int> alts;
       if(!iswupper(val))
       {
-        cur.step(val, alphabet(L"<ANY_CHAR>"));
+        alts.insert(alphabet(L"<ANY_CHAR>"));
+        alts.insert(alphabet(L"<ANY_LOWER>"));
       }
       else
       {
-        cur.step(val, towlower(val), alphabet(L"<ANY_CHAR>"));
+        alts.insert(alphabet(L"<ANY_CHAR>"));
+        alts.insert(alphabet(L"<ANY_UPPER>"));
+        alts.insert(towlower(val));
       }
+      cur.step(val, alts);
 
     }
   }
@@ -273,7 +272,7 @@ LRXProcessor::recognisePattern(const wstring lu, const wstring op)
   return false;
 }
 
-
+/*
 void
 LRXProcessor::processFlush(FILE *output,
                            map<int, wstring > &sl,
@@ -469,8 +468,9 @@ LRXProcessor::processFlush(FILE *output,
     }
   }
 }
+*/
 
-
+/*
 void
 LRXProcessor::process(FILE *input, FILE *output)
 {
@@ -782,9 +782,10 @@ LRXProcessor::process(FILE *input, FILE *output)
 
   fwprintf(output, L"%S", blanks[pos].c_str());
 }
+*/
 
 void
-LRXProcessor::processME(FILE *input, FILE *output)
+LRXProcessor::process(FILE *input, FILE *output)
 {
   bool isEscaped = false;
 
@@ -793,21 +794,27 @@ LRXProcessor::processME(FILE *input, FILE *output)
   map<int, wstring > blanks; // map of the superblanks
 
   map<int, map<wstring, double> > scores; //
+  map<int, map<wstring, OpType> > operations;
 
   vector<State*> alive_states ;
   alive_states.push_back(new State(*initial_state));
 
   int val = 0;
-  while((val = fgetwc_unlocked(input)) != EOF)
+  while((val = fgetwc_unlocked(input)) != EOF && val != WEOF)
   {
 
     if(nullFlush && val == L'\0')
     {
-      processFlushME(output, sl, tl, blanks, scores);
+      processFlush(output, sl, tl, blanks, scores, operations);
+      fwprintf(output, L"%S", blanks[pos].c_str());
       pos = 0;
       tl.clear();
       sl.clear();
       blanks.clear();
+      scores.clear();
+      operations.clear();
+      alive_states.clear();
+      alive_states.push_back(new State(*initial_state));
 
       fputwc_unlocked(val, output);
       fflush(output);
@@ -827,7 +834,7 @@ LRXProcessor::processME(FILE *input, FILE *output)
       // Read in target equivalences
       wstring trad = L"";
       val = fgetwc_unlocked(input);
-      while(val != L'$' && val != EOF)
+      while(val != L'$' && val != EOF && val != WEOF)
       {
         if(val != L'$')
         {
@@ -844,9 +851,9 @@ LRXProcessor::processME(FILE *input, FILE *output)
 
       if(debugMode)
       {
-        for(vector<wstring>::iterator it = tl[pos].begin(); it != tl[pos].end(); it++)
+        for(auto& it : tl[pos])
         {
-          fwprintf(stderr, L"trad[%d]: %S\n", pos, it->c_str());
+          fwprintf(stderr, L"trad[%d]: %S\n", pos, it.c_str());
         }
       }
     }
@@ -861,9 +868,9 @@ LRXProcessor::processME(FILE *input, FILE *output)
         vector<State *> new_states; // TODO: Can we avoid the State-copying here?
         // \forall s \in A
         set<wstring> seen_ids;
-        for(vector<State *>::const_iterator it = alive_states.begin(); it != alive_states.end(); it++)
+        for(auto& it : alive_states)
         {
-          State s = **it;
+          State s = *it;
           // \IF \exists c \in Q : \delta(s, sent[i]) = c
           s.step(alphabet(L"<$>"));
 
@@ -888,13 +895,11 @@ LRXProcessor::processME(FILE *input, FILE *output)
             set<pair<wstring, vector<wstring>>> outpaths;
             outpaths = s.filterFinalsLRX(anfinals, alphabet, escaped_chars, false, false, 0);
 
-            set<pair<wstring, vector<wstring>>>::iterator it;
-            for (it = outpaths.begin(); it != outpaths.end(); it++)
+            for (auto& it : outpaths)
             {
               vector<State> reached;
-
-              vector<wstring> path = (*it).second;
-              wstring id = (*it).first;
+              vector<wstring> path = it.second;
+              wstring id = it.first;
 
               if (seen_ids.find(id) != seen_ids.end())
               {
@@ -908,22 +913,28 @@ LRXProcessor::processME(FILE *input, FILE *output)
               {
                 fwprintf(stderr, L"id:      %S: (lambda: %.5f)\n", id.c_str(), weights[id.c_str()]);
               }
-              for (vector<wstring>::iterator it2 = path.begin(); it2 != path.end(); it2++)
+              for (auto& it2 : path)
               {
                 if (debugMode)
                 {
-                  fwprintf(stderr, L"op:        %S\n", it2->c_str());
+                  fwprintf(stderr, L"op:        %S\n", it2.c_str());
                 }
-                if (*it2 != LRX_PROCESSOR_TAG_SKIP)
+                if (it2 != LRX_PROCESSOR_TAG_SKIP)
                 {
-                  if (scores[j].count(*it2) == 0)
+                  if (scores[j].count(it2) == 0)
                   {
-                    scores[j][*it2] = 0.0;
+                    scores[j][it2] = 0.0;
                   }
-                  scores[j][*it2] += weights[id.c_str()];
+                  scores[j][it2] += weights[id.c_str()];
                   if (debugMode)
                   {
-                    fwprintf(stderr, L"#[%d]SCORE %.5f / %S\n", j, scores[j][*it2], it2->c_str());
+                    fwprintf(stderr, L"#[%d]SCORE %.5f / %S\n", j, scores[j][it2], it2.c_str());
+                  }
+                  if(it2.at(0) == L'<' && it2.at(1) == L'r') {
+                    operations[j][it2] = Remove;
+                  }
+                  else {
+                    operations[j][it2] = Select;
                   }
                 }
                 j++;
@@ -934,7 +945,7 @@ LRXProcessor::processME(FILE *input, FILE *output)
         }
         alive_states.swap(new_states);
         alive_states.push_back(new State(*initial_state));
-        for (State *s : new_states) {
+        for (auto& s : new_states) {
           if (s != initial_state) {
             delete s;
           }
@@ -943,9 +954,9 @@ LRXProcessor::processME(FILE *input, FILE *output)
         if (debugMode)
         {
           fwprintf(stderr, L"seen:");
-          for (set<wstring>::iterator it = seen_ids.begin(); it != seen_ids.end(); it++)
+          for (auto& it : seen_ids)
           {
-            fwprintf(stderr, L" %S ", it->c_str());
+            fwprintf(stderr, L" %S ", it.c_str());
           }
           fwprintf(stderr, L"\n");
           fwprintf(stderr, L"#CURRENT_ALIVE: %d\n", alive_states.size());
@@ -964,85 +975,7 @@ LRXProcessor::processME(FILE *input, FILE *output)
 
 
         // Here we actually apply the rules that we've matched
-
-        unsigned int spos = 0;
-        for(spos = 0; spos <= pos; spos++)
-        {
-          if(sl[spos] == L"")
-          {
-            continue;
-          }
-
-          fwprintf(output, L"%S^%S/", blanks[spos].c_str(), sl[spos].c_str());
-
-          vector<wstring>::iterator ti;
-          vector<wstring>::iterator penum = tl[spos].end(); penum--;
-
-          if(tl[spos].size() > 1)
-          {
-            //--
-            double l_max = 0.0;
-            wstring ti_max;
-            for(ti = tl[spos].begin(); ti != tl[spos].end(); ti++)
-            {
-
-                map<wstring, double>::iterator si;
-                for(si = scores[spos].begin(); si != scores[spos].end(); si++)
-                {
-                  if(debugMode)
-                  {
-                    fwprintf(stderr, L">>> %d -> %S -> %.5f\n", spos, si->first.c_str(), si->second);
-                  }
-                  bool matched = false;
-                  matched = recognisePattern(*ti, si->first);
-                  if(si->second > l_max && matched)
-                  {
-                    l_max = si->second;
-                    ti_max = *ti;
-                  }
-                }
-            }
-
-            if(l_max > 0.0)  // If we actually got a winner
-            {
-              if(traceMode || debugMode)
-              {
-                //fwprintf(stderr, L"%d: %d: %S -> %S (%d)\n", lineno, spos, sl[spos].c_str(), ti->c_str(), scores[spos].size());
-                //fwprintf(stderr, L"MAX: %.5f = %S\n", l_max, ti_max.c_str());
-                fwprintf(stderr, L"%d:SELECT:%.5f:%S:%S\n", lineno, l_max, sl[spos].c_str(), ti_max.c_str());
-              }
-              fwprintf(output, L"%S", ti_max.c_str());
-            }
-            else
-            {
-              for(ti = tl[spos].begin(); ti != tl[spos].end(); ti++)
-              {
-                fwprintf(output, L"%S", ti->c_str());
-                if(ti != penum)
-                {
-                  fwprintf(output, L"/");
-                }
-              }
-            }
-          }
-          else
-          {
-            for(ti = tl[spos].begin(); ti != tl[spos].end(); ti++)
-            {
-              fwprintf(output, L"%S", ti->c_str());
-              if(ti != penum)
-              {
-                fwprintf(output, L"/");
-              }
-            }
-          }
-
-          fwprintf(output, L"$");
-          if(debugMode)
-          {
-            fwprintf(output, L"%d", spos);
-          }
-        }
+        processFlush(output, sl, tl, blanks, scores, operations);
 
         pos = 0;
         tl.clear();
@@ -1087,10 +1020,9 @@ LRXProcessor::processME(FILE *input, FILE *output)
       }
 
       wstring res = L"";
-      for(vector<State*>::const_iterator it = alive_states.begin(); it != alive_states.end(); it++)
+      for(auto& s : alive_states)
       {
         res = L"";
-        State* s = *it;
         if(val < 0)
         {
           alphabet.getSymbol(res, val,  false);
@@ -1102,11 +1034,23 @@ LRXProcessor::processME(FILE *input, FILE *output)
         }
         else
         {
+
+          set<int> alts;
+          alts.insert(alphabet(L"<ANY_CHAR>"));
+          if(iswupper(val))
+          {
+            alts.insert(towlower(val));
+            alts.insert(alphabet(L"<ANY_UPPER>"));
+          }
+          else
+          {
+            alts.insert(alphabet(L"<ANY_LOWER>"));
+          }
           if(debugMode)
           {
-            fwprintf(stderr, L"  step: %C\n", val);
+            fwprintf(stderr, L"  step: %C [alts: %d]\n", val, alts.size());
           }
-          s->step_case(val, alphabet(L"<ANY_CHAR>"), false);
+          s->step(val, alts);
         }
       }
       // In the middle of a word, don't push initial state here cf. https://github.com/apertium/apertium-lex-tools/issues/19
@@ -1139,15 +1083,23 @@ LRXProcessor::processME(FILE *input, FILE *output)
 
   }
 
-  processFlushME(output, sl, tl, blanks, scores);
+  processFlush(output, sl, tl, blanks, scores, operations);
+  fwprintf(output, L"%S", blanks[pos].c_str());
 }
 
 void
-LRXProcessor::processFlushME(FILE *output,
+LRXProcessor::processFlush(FILE *output,
                              map<int, wstring > &sl,
                              map<int, vector<wstring> > &tl,
                              map<int, wstring > &blanks,
-                             map<int, map<wstring, double> > &scores) {
+                             map<int, map<wstring, double> > &scores,
+                             map<int, map<wstring, OpType> > &operations) {
+
+  struct ScoredMatch {
+      OpType op;
+      wstring* ti;              // matched target translation
+      double weight;
+  };
 
   unsigned int spos = 0;
   for(spos = 0; spos <= pos; spos++)
@@ -1160,42 +1112,67 @@ LRXProcessor::processFlushME(FILE *output,
     fwprintf(output, L"%S^%S/", blanks[spos].c_str(), sl[spos].c_str());
 
     vector<wstring>::iterator ti;
-    vector<wstring>::iterator penum = tl[spos].end(); penum--;
+    auto penum = tl[spos].end();
+    penum--;
 
     if(tl[spos].size() > 1)
     {
       //--
-      double l_max = 0.0;
-      wstring ti_max;
+      set<wstring*> ti_keep;
+      set<wstring*> ti_removed;
+      vector<ScoredMatch> spos_matches;
       for(ti = tl[spos].begin(); ti != tl[spos].end(); ti++)
       {
-
-        map<wstring, double>::iterator si;
-        for(si = scores[spos].begin(); si != scores[spos].end(); si++)
-        {
-          if(debugMode)
-          {
-            fwprintf(stderr, L">>> %d -> %S -> %.5f\n", spos, si->first.c_str(), si->second);
+        ti_keep.insert(&*ti);
+        for(const auto& si : scores[spos]) {
+          bool matched = recognisePattern(*ti, si.first);
+          OpType op = operations[spos][si.first];
+          if (debugMode) {
+            wstring checks = matched ? L"✔️ " : L"❎";
+            fwprintf(stderr, L"%S >>> %d -> %S -> %.5f\n", checks.c_str(), spos,
+                     si.first.c_str(), si.second);
           }
-          bool matched = false;
-          matched = recognisePattern(*ti, si->first);
-          if(si->second > l_max && matched)
-          {
-            l_max = si->second;
-            ti_max = *ti;
+          if(matched) {
+            spos_matches.push_back({ op, &*ti, si.second });
           }
         }
       }
-
-      if(l_max > 0.0)  // If we actually got a winner
+      if(!spos_matches.empty())  // If we actually got a winner
       {
-        if(traceMode || debugMode)
-        {
-          //fwprintf(stderr, L"%d: %d: %S -> %S (%d)\n", lineno, spos, sl[spos].c_str(), ti->c_str(), scores[spos].size());
-          //fwprintf(stderr, L"MAX: %.5f = %S\n", l_max, ti_max.c_str());
-          fwprintf(stderr, L"%d:SELECT:%.5f:%S:%S\n", lineno, l_max, sl[spos].c_str(), ti_max.c_str());
+        sort(spos_matches.begin(),
+             spos_matches.end(),
+             [](const auto &a, const auto &b) { return a.weight > b.weight; });
+        for (const auto &m : spos_matches) {
+          if (traceMode || debugMode) {
+            wstring op = (m.op == Select ? L"SELECT" : L"REMOVE");
+            fwprintf(
+                stderr, L"%d:%S:%.5f:%S:%d:%S\n",
+                lineno,
+                op.c_str(),
+                m.weight,
+                sl[spos].c_str(),
+                ti_keep.size(),
+                m.ti->c_str());
+          }
+          // We have to keep track of translations that have been removed so
+          // that we don't end up adding back a translation that was removed.
+          if (m.op == Select && ti_removed.find(m.ti) == ti_removed.end()) {
+            ti_keep.clear();
+            ti_keep.insert(m.ti);
+            break;
+          } else if(ti_keep.size() > 1) {
+            ti_keep.erase(m.ti);
+            ti_removed.insert(m.ti);
+          }
+        }
+        bool printed = false;
+        for(const auto& ti_max : ti_keep) {
+          if(printed) {
+            fwprintf(output, L"/");
+          }
+          fwprintf(output, L"%S", ti_max->c_str());
+          printed = true;
         }
-        fwprintf(output, L"%S", ti_max.c_str());
       }
       else
       {
@@ -1230,5 +1207,4 @@ LRXProcessor::processFlushME(FILE *output,
 
   }
 
-  fwprintf(output, L"%S", blanks[pos].c_str());
 }
